import collections

import numpy as np
from . import Misc as misc
from . import SetOps as sets
from . import VectorOps as vec_ops

__all__ = [
    "permutation_sign",
    "levi_cevita_maps",
    "levi_cevita_tensor",
    "levi_cevita3",
    "levi_cevita_dot",
    "normalize_commutators",
    "commutator_terms",
    "commutator_evaluate",
    "permutation_cycles",
    "enumerate_permutations"
]


def permutation_sign(perm, check=True):
    # essentially a swap sort on perm
    # https://stackoverflow.com/a/73511014
    parity = 1
    perm = np.asanyarray(perm)
    if check:
        perm = np.argsort(np.argsort(perm))
    else:
        perm = perm.copy()
    for i in range(len(perm)):
        while perm[i] != i: # ensure subblock is sorted
            parity *= -1
            j = perm[i]
            perm[i], perm[j] = perm[j], perm[i]
    return parity
def levi_cevita_maps(k):
    perms = sets.permutation_indices(k, k)
    signs = np.array([permutation_sign(p, check=False) for p in perms])
    return perms, signs
def levi_cevita_tensor(k, sparse=False):
    pos, vals = levi_cevita_maps(k)
    if sparse:
        from .Sparse import SparseArray
        a = SparseArray.from_data(
            (
                pos.T,
                vals
            ),
            shape=(k,)*k
        )
    else:
        a = np.zeros((k,)*k, dtype=int)
        a[tuple(pos.T)] = vals
    return a
# levi_cevita3 = levi_cevita_tensor(3)
levi_cevita3 = np.array([
    [[0, 0, 0],
     [0, 0, 1],
     [0, -1, 0]],

    [[0, 0, -1],
     [0, 0, 0],
     [1, 0, 0]],

    [[0, 1, 0],
     [-1, 0, 0],
     [0, 0, 0]]
])

def levi_cevita_dot(k, a, /, axes, shared=None):
    pos, vals = levi_cevita_maps(k)
    return vec_ops.semisparse_tensordot((tuple(pos.T), vals, (k,) * k), a, axes, shared=shared)

def _flatten_comstr(cs):
    for o in cs:
        if isinstance(o, (int, np.integer)):
            yield o
        else:
            for f in _flatten_comstr(o):
                yield f

def normalize_commutators(commutator_string):
    a, b = commutator_string
    a_int = isinstance(a, (int, np.integer))
    b_int = isinstance(b, (int, np.integer))
    if b_int and a_int:
        return [1], [commutator_string], [a, b]
    elif a_int:
        ps, bs, ts = normalize_commutators(b)
        return [-p for p in ps], [
            [b, a]
            for b in bs
        ], ts + [a]
    elif b_int:
        ps, as_, ts = normalize_commutators(a)
        return [p for p in ps], [
            [a, b]
            for a in as_
        ], ts + [b]
    else:
        aa, bb = a
        aa_int = isinstance(aa, (int, np.integer))
        bb_int = isinstance(bb, (int, np.integer))
        cc, dd = b
        cc_int = isinstance(cc, (int, np.integer))
        dd_int = isinstance(dd, (int, np.integer))

        if aa_int and bb_int:
            if cc_int and dd_int:
                forms = [
                    [[[aa, bb], cc], dd],
                    [[[dd, aa], bb], cc],
                    [[[cc, dd], aa], bb],
                    [[[bb, cc], dd], aa]
                ]
                phases = [1, 1, 1, 1]
                terms = [aa, bb, cc, dd]
                return phases, forms, terms
            else:
                pb, bs_, tb = normalize_commutators(b)
                return [-p for p in pb], [
                    [b, a]
                    for b in bs_
                ], tb + [aa, bb]
        elif cc_int and dd_int:
            pa, as_, ta = normalize_commutators(a)
            return [p for p in pa], [
                [a, b]
                for a in as_
            ], ta + [cc, dd]
        else:
            pa, as_, ta = normalize_commutators(a)
            pb, bs_, tb = normalize_commutators(b)
            if len(tb) > len(ta):
                return [
                    -p1 * p2
                    for p2 in pb
                    for p1 in pa
                ], [
                    [b, a]
                    for b in bs_
                    for a in as_
                ], tb + ta
            else:
                return [
                    p1 * p2
                    for p1 in pa
                    for p2 in pb
                ], [
                    [a, b]
                    for a in as_
                    for b in bs_
                ], ta + tb

def _setup_com_terms(full_phases, storage, i0, idx, j0, j, term):
    a,b = term
    if isinstance(a, (int, np.integer)): # simple swap, nothing more needed
        prev = storage[i0:i0+idx]
        n = idx * 2
        new = storage[i0+idx:i0+n]
        new[:, :j0] = prev[:, :j0]
        new[:, j0], new[:, j0+1:j+j0+1] = prev[:, j0+j], prev[:, j0:j0+j]
        full_phases[i0+idx:i0+n] = -full_phases[i0:i0+idx]
        j = j + 1
    else:
        idx, j = _setup_com_terms(full_phases, storage, i0, idx, j0, j, a)
        if isinstance(b, (int, np.integer)): # swap all priors
            n = idx * 2
            prev = storage[i0:i0+idx]
            new = storage[i0 + idx:i0 + n]
            new[:, :j0] = prev[:, :j0]
            new[:, j0], new[:, j0 + 1:j + j0 + 1] = prev[:, j0 + j], prev[:, j0:j0 + j]
            full_phases[i0+idx:i0+n] = -full_phases[i0:i0+idx]
            j = j + 1
        else:
            n, j1 = _setup_com_terms(full_phases, storage, i0, idx, j, 1, b)
            idx = n
            n = 2 * n
            prev = storage[i0:i0+idx]
            new = storage[i0+idx:i0+n]
            new[:, j0:j0+j1], new[:, j0+j1:j0+j1+j] = prev[:, j0+j:j0+j+j1], prev[:, j0:j0+j]
            full_phases[i0+idx:i0+n] = -full_phases[i0:i0+idx]
    return n, j

def commutator_terms(commutator_strings):
    phases, normal_forms, symbols = normalize_commutators(commutator_strings)
    storage = np.full((2**(len(symbols)-1), len(symbols)), symbols)
    full_phases = np.ones(len(storage), dtype=int)
    idx = 0
    for base_phase,term in zip(phases, normal_forms):
        nx, _ = _setup_com_terms(full_phases, storage, idx, 1, 0, 1, term)
        full_phases[idx:idx+nx] *= base_phase
        idx += nx

    return full_phases, storage

def commutator_evaluate(commutator, expansion_terms, normalized=False, direct=None, recursive=False):
    if recursive:
        a,b = commutator
        if not isinstance(a, (int, np.integer)):
            a = commutator_evaluate(a, expansion_terms, recursive=True)
        else:
            a = expansion_terms[a]
        if not isinstance(b, (int, np.integer)):
            b = commutator_evaluate(b, expansion_terms, recursive=True)
        else:
            b = expansion_terms[b]
        return a@b - b@a

    if direct is None:
        test_phases, test_terms = commutator
        direct = (
                isinstance(test_phases, (int, np.integer))
                or not isinstance(test_phases[0], (int, np.integer))
                or not (abs(test_phases[-1]) == 1 and abs(test_phases[0]) == 1)
        )
    if direct:
        terms = collections.deque()
        terms.append([(0,), commutator])
        exprs = {}
        while terms:
            idx, (a,b) = terms.pop()
            ta = exprs.pop(idx + (0,), None)
            tb = exprs.pop(idx + (1,), None)
            if ta is None and isinstance(a, (int, np.integer)):
                ta = expansion_terms[a]
            if tb is None and isinstance(b, (int, np.integer)):
                tb = expansion_terms[b]
            if ta is not None and tb is not None:
                exprs[idx] = ta @ tb - tb @ ta
            else:
                terms.append([idx,(None, None)]) # by the time this comes back up we expect those to be filled in
                if tb is None:
                    terms.append([idx+(1,), b])
                else:
                    exprs[idx+(1,)] = tb
                if ta is None:
                    terms.append([idx+(0,), a])
                else:
                    exprs[idx+(0,)] = ta
        return exprs[(0,)]
    else:
        if not normalized:
            commutator = commutator_terms(commutator)
        phases, terms = commutator
        comm = 0
        for p,t in zip(phases, terms):
            res = expansion_terms[t[0]]
            for i in t[1:]:
                res = res @ expansion_terms[i]
            if p < 0:
                comm -= res
            else:
                comm += res
        return comm


def permutation_cycles(perms, return_groups=False):
    # since cycles have no set size, this can't be efficiently vectorized
    # a much better candidate for speeding up with numba if we need it
    perms = np.asanyarray(perms)
    if return_groups and perms.ndim > 2:
        raise ValueError(
            f"can't destructure permutations of shape {perms.shape} into groups, "
            "try `return_groups=False` to just get labels"
        )
    base_shape = perms.shape[:-1]
    perms = perms.reshape(-1, perms.shape[-1])

    labels = np.full(perms.shape, -1, dtype=int)
    group_indicators = np.zeros(len(perms), dtype=int)
    for i in range(perms.shape[-1]):
        mask = np.where(labels[:, i] < 0)
        if len(mask) == 0 or len(mask[0]) == 0:
            continue

        mask = mask[0]
        start = new = perms[mask, i]
        x = group_indicators[mask,]
        group_indicators[mask,] += 1
        for _ in range(perms.shape[-1]):
            labels[mask, new] = x
            new = perms[mask, new]
            checks = np.where(start == new)
            if len(checks) > 0 and len(checks[0]) > 0:
                checks = checks[0]
                if len(checks) == len(mask): break
                rem = np.delete(np.arange(len(mask)), checks)
                mask = mask[rem,]
                x = x[rem,]
                new = new[rem,]
                start = start[rem,]
            labels[mask, new] = x

    labels = labels.reshape(base_shape + (labels.shape[-1],))
    if return_groups:
        if labels.ndim == 1:
            groups = sets.group_by(np.arange(len(labels)), labels)[0][1]
            return [
                perms[0][g] for g in groups
            ]
        elif labels.ndim == 2:
            cycle_lists = []
            for l, p in zip(labels, perms):
                groups = sets.group_by(np.arange(len(l)), l)[0][1]
                cycle_lists.append([
                    p[g] for g in groups
                ])
            return cycle_lists

    return labels
def permutation_from_cycles(cycle):
    base_concat = np.concatenate(cycle, axis=-1)
    subconcat = np.concatenate([np.roll(p, 1, axis=-1) for p in cycle])
    sort = np.argsort(base_concat, axis=-1)
    return subconcat[sort]

def compute_cycle_orders(perms):
    perms = np.asanyarray(perms)
    if perms.ndim == 1:
        groups = permutation_cycles(perms, return_groups=True)
        cycle_orders = np.prod([len(g) for g in groups], dtype=int)
    else:
        cycles = permutation_cycles(perms, return_groups=False).view('uint64')  # want underflow
        max_cycle_size = np.max(cycles)
        cycle_orders = np.ones(perms.shape[:-1], dtype=int)
        cycles = np.sort(cycles)
        for i in range(max_cycle_size):
            subcounts = sets.fast_first_nonzero(cycles)
            subcounts[subcounts < 0] = perms.shape[-1]
            subcounts[subcounts == 0] = 1
            cycle_orders *= subcounts
            cycles = np.sort(cycles - 1)  # yes I should do this better but...
    return cycle_orders

def enumerate_permutations(perm, cycle_orders=None):
    perm = np.asanyarray(perm)
    if perm.ndim == 1:
        if cycle_orders is None:
            groups = permutation_cycles(perm, return_groups=True)
            cycle_orders = np.prod([len(g) for g in groups], dtype=int)
        perms = np.empty((cycle_orders, perm.shape[-1]), dtype=perm.dtype)
        perms[0, :] = perm
        for i in range(1, cycle_orders):
            perms[i, :] = perms[i-1, :][perm,]
        return perms
    elif perm.ndim > 2:
        raise NotImplementedError("too annoying to do ndim > 2")
    else:
        if cycle_orders is None:
            cycle_orders = compute_cycle_orders(perm)
        return [
            enumerate_permutations(p, c)
            for p,c in zip(perm, cycle_orders)
        ]





import itertools
from xml.etree import ElementTree
import weakref, numpy as np, copy, textwrap, inspect

__all__ = [
    "HTML",
    "CSS",
    "ContentXML"
]

from ...Misc import mixedmethod
from .Enums import Options
from .WidgetTools import frozendict

class CSS:
    """
    Defines a holder for CSS properties
    """
    def __init__(self, *selectors, **props):
        self.selectors = selectors
        self.props = self.canonicalize_props(props)
    known_properties = set(o.value for o in Options)
    @classmethod
    def construct(cls,
                  *selectors,
                  background=None,
                  background_attachment=None,
                  background_color=None,
                  background_image=None,
                  background_position=None,
                  background_repeat=None,
                  border=None,
                  border_bottom=None,
                  border_bottom_color=None,
                  border_bottom_style=None,
                  border_bottom_width=None,
                  border_color=None,
                  border_left=None,
                  border_left_color=None,
                  border_left_style=None,
                  border_left_width=None,
                  border_right=None,
                  border_right_color=None,
                  border_right_style=None,
                  border_right_width=None,
                  border_style=None,
                  border_top=None,
                  border_top_color=None,
                  border_top_style=None,
                  border_top_width=None,
                  border_width=None,
                  clear=None,
                  clip=None,
                  color=None,
                  cursor=None,
                  display=None,
                  filter=None,
                  float=None,
                  font=None,
                  font_family=None,
                  font_size=None,
                  font_variant=None,
                  font_weight=None,
                  height=None,
                  left=None,
                  letter_spacing=None,
                  line_height=None,
                  list_style=None,
                  list_style_image=None,
                  list_style_position=None,
                  list_style_type=None,
                  margin=None,
                  margin_bottom=None,
                  margin_left=None,
                  margin_right=None,
                  margin_top=None,
                  overflow=None,
                  padding=None,
                  padding_bottom=None,
                  padding_left=None,
                  padding_right=None,
                  padding_top=None,
                  page_break_after=None,
                  page_break_before=None,
                  position=None,
                  text_align=None,
                  text_decoration=None,
                  text_indent=None,
                  text_transform=None,
                  top=None,
                  vertical_align=None,
                  visibility=None,
                  width=None,
                  z_index=None,
                  **props
                  ):
        """
        Provides a convenience constructor for systems with autocompletions

        :param selectors:
        :type selectors:
        :param background:
        :type background:
        :param background_attachment:
        :type background_attachment:
        :param background_color:
        :type background_color:
        :param background_image:
        :type background_image:
        :param background_position:
        :type background_position:
        :param background_repeat:
        :type background_repeat:
        :param border:
        :type border:
        :param border_bottom:
        :type border_bottom:
        :param border_bottom_color:
        :type border_bottom_color:
        :param border_bottom_style:
        :type border_bottom_style:
        :param border_bottom_width:
        :type border_bottom_width:
        :param border_color:
        :type border_color:
        :param border_left:
        :type border_left:
        :param border_left_color:
        :type border_left_color:
        :param border_left_style:
        :type border_left_style:
        :param border_left_width:
        :type border_left_width:
        :param border_right:
        :type border_right:
        :param border_right_color:
        :type border_right_color:
        :param border_right_style:
        :type border_right_style:
        :param border_right_width:
        :type border_right_width:
        :param border_style:
        :type border_style:
        :param border_top:
        :type border_top:
        :param border_top_color:
        :type border_top_color:
        :param border_top_style:
        :type border_top_style:
        :param border_top_width:
        :type border_top_width:
        :param border_width:
        :type border_width:
        :param clear:
        :type clear:
        :param clip:
        :type clip:
        :param color:
        :type color:
        :param cursor:
        :type cursor:
        :param display:
        :type display:
        :param filter:
        :type filter:
        :param float:
        :type float:
        :param font:
        :type font:
        :param font_family:
        :type font_family:
        :param font_size:
        :type font_size:
        :param font_variant:
        :type font_variant:
        :param font_weight:
        :type font_weight:
        :param height:
        :type height:
        :param left:
        :type left:
        :param letter_spacing:
        :type letter_spacing:
        :param line_height:
        :type line_height:
        :param list_style:
        :type list_style:
        :param list_style_image:
        :type list_style_image:
        :param list_style_position:
        :type list_style_position:
        :param list_style_type:
        :type list_style_type:
        :param margin:
        :type margin:
        :param margin_bottom:
        :type margin_bottom:
        :param margin_left:
        :type margin_left:
        :param margin_right:
        :type margin_right:
        :param margin_top:
        :type margin_top:
        :param overflow:
        :type overflow:
        :param padding:
        :type padding:
        :param padding_bottom:
        :type padding_bottom:
        :param padding_left:
        :type padding_left:
        :param padding_right:
        :type padding_right:
        :param padding_top:
        :type padding_top:
        :param page_break_after:
        :type page_break_after:
        :param page_break_before:
        :type page_break_before:
        :param position:
        :type position:
        :param text_align:
        :type text_align:
        :param text_decoration:
        :type text_decoration:
        :param text_indent:
        :type text_indent:
        :param text_transform:
        :type text_transform:
        :param top:
        :type top:
        :param vertical_align:
        :type vertical_align:
        :param visibility:
        :type visibility:
        :param width:
        :type width:
        :param z_index:
        :type z_index:
        :param props:
        :type props:
        :return:
        :rtype:
        """
        common_props = dict(
            background=background,
            background_attachment=background_attachment,
            background_color=background_color,
            background_image=background_image,
            background_position=background_position,
            background_repeat=background_repeat,
            border=border,
            border_bottom=border_bottom,
            border_bottom_color=border_bottom_color,
            border_bottom_style=border_bottom_style,
            border_bottom_width=border_bottom_width,
            border_color=border_color,
            border_left=border_left,
            border_left_color=border_left_color,
            border_left_style=border_left_style,
            border_left_width=border_left_width,
            border_right=border_right,
            border_right_color=border_right_color,
            border_right_style=border_right_style,
            border_right_width=border_right_width,
            border_style=border_style,
            border_top=border_top,
            border_top_color=border_top_color,
            border_top_style=border_top_style,
            border_top_width=border_top_width,
            border_width=border_width,
            clear=clear,
            clip=clip,
            color=color,
            cursor=cursor,
            display=display,
            filter=filter,
            float=float,
            font=font,
            font_family=font_family,
            font_size=font_size,
            font_variant=font_variant,
            font_weight=font_weight,
            height=height,
            left=left,
            letter_spacing=letter_spacing,
            line_height=line_height,
            list_style=list_style,
            list_style_image=list_style_image,
            list_style_position=list_style_position,
            list_style_type=list_style_type,
            margin=margin,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            overflow=overflow,
            padding=padding,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            page_break_after=page_break_after,
            page_break_before=page_break_before,
            position=position,
            text_align=text_align,
            text_decoration=text_decoration,
            text_indent=text_indent,
            text_transform=text_transform,
            top=top,
            vertical_align=vertical_align,
            visibility=visibility,
            width=width,
            z_index=z_index
        )
        for k, v in common_props.items():
            if v is not None:
                props[k] = v
        return cls(*selectors, **props)
    @classmethod
    def canonicalize_props(cls, props):
        return {k.replace("_", "-"):v for k,v in props.items()}
    @classmethod
    def parse(cls, sty):
        header = sty.split("{", 1)[0]
        if len(header) == len(sty): # inline styles
            chunks = [x.strip() for x in sty.split(";")]
            splits = [x.split(":") for x in chunks if len(x) > 0]
            return cls(**{k.strip(): v.strip() for k, v in splits})
        else:
            splits = [x.split("{") for x in sty.split("}")]
            styles = []
            for key, vals in splits:
                key = [x.strip() for x in key.split(",")]
                chunks = [x.strip() for x in vals.split(";")]
                pairs = [x.split(":") for x in chunks if len(x) > 0]
                styles.append(
                    cls(*key, **{k.strip(): v.strip() for k, v in pairs})
                )
                return styles

    def tostring(self):
        if len(self.selectors) > 0:
            return "{sel} {{\n  {body}\n}}".format(
                sel=",".join(self.selectors),
                body="\n  ".join("{k}:{v};".format(k=k,v=v) for k,v in self.props.items())
            )
        else:
            return " ".join("{k}:{v};".format(k=k,v=v) for k,v in self.props.items())

class HTMLManager:
    @classmethod
    def manage_class(kls, cls):
        if cls is None:
            cls = []
        elif hasattr(cls, 'tostring'):
            cls = cls.tostring
        if isinstance(cls, str):
            cls = cls.split()
        else:
            try:
                iter(cls)
            except TypeError:
                cls = str(cls).split()
        return list(cls)

    @classmethod
    def manage_styles(cls, styles):
        if hasattr(styles, 'items'):
            styles = CSS(**styles)
        elif isinstance(styles, str):
            styles = CSS.parse(styles)
        return styles

    keyword_replacements = {
        'cls': 'class',
        'use_for': 'for',
        'custom_type': 'is'
    }

    @classmethod
    def clean_key(cls, k):
        if k in cls.keyword_replacements:
            return cls.keyword_replacements[k]
        else:
            return k.replace("_", "-")

    @classmethod
    def sanitize_value(cls, val):
        if isinstance(val, np.integer):
            val = int(val)
        elif isinstance(val, np.floating):
            val = float(val)
        return val

    @classmethod
    def manage_attrs(cls, attrs, sanitize=True):
        for k, v in cls.keyword_replacements.items():
            if k in attrs:
                attrs[v] = attrs[k]
                del attrs[k]
        attrs = {k.replace("_", "-"): v for k, v in attrs.items()}
        if sanitize:
            attrs = {k: cls.sanitize_value(v) for k, v in attrs.items()}
        return attrs

    @classmethod
    def extract_styles(cls, attrs, style_props=None, ignored_styles=None):
        if style_props is None:
            style_props = CSS.known_properties
        if ignored_styles is not None:
            style_props = style_props - set(ignored_styles)
        styles = {}
        for k, v in tuple(attrs.items()):
            if k in style_props:
                styles[k] = v
                del attrs[k]
        return styles, attrs

    class ElementModifier:
        def __init__(self, my_el, copy=False):
            self.el = my_el
            self.needs_copy = copy
            self._parents = None
            self._tree_cache = None
        def modify(self):
            if self.needs_copy:
                el = self.el.copy()
            else:
                el = self.el
            return el
        def tostring(self):
            return self.modify().tostring()
        def _repr_html_(self):
            return self.tostring()
        def copy(self):
            import copy
            new = copy.copy(self)
            new.el = new.el.copy()
        def add_class(self, *cls, copy=True):
            return self.el.context.ClassAdder(self, cls=cls, copy=copy)
        def remove_class(self, *cls, copy=True):
            return self.el.ClassRemover(self, cls=cls, copy=copy)
        def add_styles(self, copy=True, **sty):
            return self.el.StyleAdder(self, copy=copy, **sty)
    class ClassAdder(ElementModifier):
        cls = None
        def __init__(self, el, cls=None, copy=True):
            if cls is None:
                cls = self.cls
            if isinstance(cls, str):
                cls = cls.split()
            self.cls = cls
            super().__init__(el, copy=copy)
        def modify(self):
            if hasattr(self.el, 'modify'):
                el = self.el.modify()
            else:
                if self.needs_copy:
                    el = self.el.copy()
                else:
                    el = self.el
            if 'class' in el.attrs:
                if isinstance(el['class'], str):
                    el.make_class_list()
                class_list = list(el['class'])
                for cls in self.cls:
                    cls = str(cls)
                    if cls not in class_list:
                        class_list.append(cls)
                el['class'] = tuple(class_list)
            else:
                el['class'] = self.cls
            return el
        def __repr__(self):
            return "{}({}, {})".format(type(self).__name__, self.el, self.cls)
    class ClassRemover(ElementModifier):
        cls = None
        def __init__(self, el, cls=None, copy=True):
            if cls is None:
                cls = self.cls
            if isinstance(cls, str):
                cls = cls.split()
            self.cls = cls
            super().__init__(el, copy=copy)
        def modify(self):
            if hasattr(self.el, 'modify'):
                el = self.el.modify()
            else:
                if self.needs_copy:
                    el = self.el.copy()
                else:
                    el = self.el
            if 'class' in el.attrs:
                if isinstance(el['class'], str):
                    el.make_class_list()
                class_list = list(el['class'])
                for cls in self.cls:
                    cls = str(cls)
                    try:
                        class_list.remove(cls)
                    except ValueError:
                        pass
                el['class'] = tuple(class_list)
            else:
                el['class'] = self.cls
            return el
        def __repr__(self):
            return "{}({}, {})".format(type(self).__name__, self.el, self.cls)
    class StyleAdder(ElementModifier):
        def __init__(self, el, copy=True, **styles):
            self.styles = styles
            super().__init__(el, copy=copy)
        def modify(self):
            if hasattr(self.el, 'modify'):
                el = self.el.modify()
            else:
                if self.needs_copy:
                    el = self.el.copy()
                else:
                    el = self.el

            if 'style' in el.attrs:
                style = el.attrs['style']
                if isinstance(style, str):
                    style = CSS.parse(style)
                else:
                    style = style.copy()
                style.props = dict(style.props, **self.styles)
                el.attrs['style'] = style
            else:
                el.attrs['style'] = CSS(**self.styles)
            return el
        def __repr__(self):
            return "{}({}, {})".format(type(self).__name__, self.el, self.styles)

    class StyleRemover(ElementModifier):
        def __init__(self, el, *styles, copy=True):
            self.styles = styles
            super().__init__(el, copy=copy)
        def modify(self):
            if hasattr(self.el, 'modify'):
                el = self.el.modify()
            else:
                if self.needs_copy:
                    el = self.el.copy()
                else:
                    el = self.el

            if 'style' in el.attrs:
                style = el.attrs['style']
                if isinstance(style, str):
                    style = CSS.parse(style)
                else:
                    style = style.copy()
                for k in self.styles:
                    if k in style.props:
                        del style.props[k]
                el.attrs['style'] = style
            # else:
            #     el.attrs['style'] = CSS(**self.styles)
            return el
        def __repr__(self):
            return "{}({}, {})".format(type(self).__name__, self.el, self.styles)

    @classmethod
    def xml_to_json(cls, tree:ElementTree.Element):
        children = []
        node = dict(tag=tree.tag, body=tree.text, tail=tree.tail, children=children, attrs=tree.attrib)
        for child in tree.getchildren():
            children.append(cls.xml_to_json(child))
        return node

class XMLBase:

    class ElementBase:
        ...

    @classmethod
    def find_globals(cls):
        for frame in inspect.stack(1):
            globs = frame.frame.f_globals
            if globs['__name__'] == '__main__':
                return globs
        else:
            return inspect.stack(1)[1].frame.f_globals

    @classmethod
    def expose(cls, globs=None):
        if globs is None:
            globs = cls.find_globals()
        for x in cls.get_class_map().values():
            globs[x.__name__] = x

    _cls_map = None
    @classmethod
    def get_class_map(cls):
        if cls._cls_map is None:
            cls._cls_map = {}
            for v in cls.__dict__.values():
                if isinstance(v, type) and hasattr(v, 'tag'):
                    cls._cls_map[v.tag] = v
        return cls._cls_map

    base_element = None
    @classmethod
    def convert(cls, etree:ElementTree.Element, strip=True, converter=None, **extra_attrs):
        import copy

        if converter is None:
            converter = cls.convert
        children = []
        for x in etree:
            if x.tail is not None:
                x = copy.copy(x)
                t = x.tail
                x.tail = None
                children.append(converter(x, strip=strip))
                children.append(t)
            else:
                children.append(converter(x, strip=strip))
        text = etree.text
        if text is not None:
            if isinstance(text, str):
                text = [text]
        else:
            text = []
        tail = etree.tail
        if tail is not None:
            if isinstance(tail, str):
                tail = [tail]
        else:
            tail = []
        tag = etree.tag

        elems = (
                [t.strip("\n") if strip else t for t in text]
                + children
                + [t.strip("\n") if strip else t for t in tail]
        )
        if strip:
            elems = [e for e in elems if not isinstance(e, str) or len(e) > 0]

        map = cls.get_class_map()
        try:
            tag_class = map[tag]
        except KeyError:
            tag_class = lambda *es,**ats:cls.base_element(tag, *es, **ats)

        attrs = {} if etree.attrib is None else etree.attrib

        return tag_class(*elems, **dict(extra_attrs, **attrs))

    @classmethod
    def parse(cls, str, strict=True, strip=True, fallback=None, converter=None):
        if strict:
            etree = ElementTree.fromstring(str)
        else:
            try:
                etree = ElementTree.fromstring(str)
            except ElementTree.ParseError as e:
                # print('no element found' in e.args[0])
                if 'junk after document element' in e.args[0]:
                    try:
                        return cls.parse('<div>\n\n'+str+'\n\n</div>', strict=True, strip=strip, fallback=fallback, converter=converter)
                    except ElementTree.ParseError:
                        if fallback is None:
                            fallback = HTML.Span
                        return fallback(str)
                if fallback is None:
                    fallback = HTML.Span
                return fallback(str)

        if converter is None:
            converter = cls.convert

        return converter(etree, strip=strip)


class HTML(XMLBase):
    """
    A namespace for holding various HTML attributes
    """

    class XMLElement(XMLBase.ElementBase):
        """
        Convenience API for ElementTree
        """

        ignored_styles = None
        style_props = None
        context = HTMLManager

        def __init__(self, tag, *elems, on_update=None, style=None, activator=None, **attrs):
            self.tag = tag
            self._elems = [
                self.context.sanitize_value(v)
                for v in (
                    elems[0]
                        if len(elems) == 1 and isinstance(elems[0], (list, tuple)) else
                    elems
                )
            ]
            self._elem_view = None
            attrs = self.context.manage_attrs(attrs)
            extra_styles, attrs = self.context.extract_styles(attrs, style_props=self.style_props, ignored_styles=self.ignored_styles)
            if style is not None:
                style = self.context.manage_styles(style).props
                for k,v in extra_styles.items():
                    if k in style:
                        raise ValueError("got style {} specified in two different locations".format(k))
                    style[k] = v
            else:
                style = extra_styles
            if len(style) > 0:
                attrs['style'] = style
            self._attrs = attrs
            self._attr_view = None
            self._parents = weakref.WeakSet()
            self._tree_cache = None
            self._json_cache = None
            self._on_update_callbacks = self._canonicalize_callback_dict(on_update)
            self.activator = activator
        class _update_callbacks:
            """
            Simple set of callbacks both weakly keyed and default
            """
            def __init__(self, base_callbacks, weak_callbacks):
                self.base_callbacks = base_callbacks
                self.weak_callbacks = weak_callbacks
            @classmethod
            def from_raw(cls, data):
                if data is None:
                    base = {}
                    weak = None
                elif isinstance(data, dict):
                    if all(k is None or isinstance(k, str) for k in data):
                        base = data
                        weak = None
                    else:
                        base = data.get(None, {})
                        if len(data) > 1:
                            weak = weakref.WeakKeyDictionary(
                                {
                                    k: {None: x if isinstance(x, list) else [x]} if not isinstance(x, dict) else x
                                    for k, x in data.items()
                                    if k is not None
                                }
                            )
                        else:
                            weak = None
                elif isinstance(data, weakref.WeakKeyDictionary):
                    base = {}
                    weak = data
                else:
                    base = {None: [data]}
                    weak = None
                return cls(base, weak)
            def items(self):
                if self.weak_callbacks is not None:
                    for k,v in self.weak_callbacks.items():
                        yield k,v
                yield None,self.base_callbacks
            def __contains__(self, item):
                if item is None:
                    return True
                elif self.weak_callbacks is None:
                    return False
                else:
                    return item in self.weak_callbacks
            def __setitem__(self, key, value):
                if key is None:
                    self.base_callbacks = value
                else:
                    if self.weak_callbacks is None: self.weak_callbacks = weakref.WeakKeyDictionary()
                    self.weak_callbacks[key] = value
            def get(self, item, default):
                if item is None:
                    return self.base_callbacks
                else:
                    if self.weak_callbacks is None:
                        return default
                    else:
                        return self.weak_callbacks.get(item, default)
            def __getitem__(self, item):
                if item is None:
                    return self.base_callbacks
                else:
                    if self.weak_callbacks is None:
                        raise KeyError("key {} not found".format(item))
                    else:
                        return self.weak_callbacks[item]
        def _canonicalize_callback_dict(self, on_update):
            if not isinstance(on_update, self._update_callbacks):
                on_update = self._update_callbacks.from_raw(on_update)
            return on_update
        def on_update(self, key, new_value, old_value, subkey=None):
            for registrant, callback_dict in self._on_update_callbacks.items():
                for f in callback_dict.get(key, []) + callback_dict.get(None, []):
                    sentinel = f(self, key, new_value, old_value, registrant, subkey)
                    # TODO: handle breaks from the sentinel
        def update_callbacks(self, key=None, registrant=None):
            return self._on_update_callbacks.get(registrant, {}).get(key, [])
        def add_update_callback(self, callback, key=None, registrant=None):
            if registrant not in self._on_update_callbacks: self._on_update_callbacks[registrant] = {}
            if key not in self._on_update_callbacks[registrant]: self._on_update_callbacks[registrant][key] = []
            self._on_update_callbacks[registrant][key].append(callback)
        def remove_update_callback(self, callback, key=None, registrant=None):
            if registrant not in self._on_update_callbacks: self._on_update_callbacks[registrant] = {}
            if key not in self._on_update_callbacks[registrant]: self._on_update_callbacks[registrant][key] = []
            self._on_update_callbacks[registrant][key].remove(callback)

        def __call__(self, *elems, **kwargs):
            return type(self)(
                self.tag,
                self._elems + list(elems),
                activator=self.activator,
                on_update=self.on_update,
                **dict(self.attrs, **kwargs)
            )
        @property
        def attrs(self):
            if self._attr_view is None:
                self._attr_view = frozendict(self._attrs)
            return self._attr_view
        @attrs.setter
        def attrs(self, attrs):
            old_attrs = self.attrs
            self._attrs = self.context.manage_attrs(attrs)
            self._attr_view = None
            self.invalidate_cache()
            self.on_update('attributes', attrs, old_attrs)
        @property
        def elems(self):
            if self._elem_view is None:
                self._elem_view = tuple(str(x) if isinstance(x, (int, float, bool)) else x for x in self._elems)
            return self._elem_view
        @elems.setter
        def elems(self, elems):
            self.set_elems(elems)
        def set_elems(self, elems):
            old_elems = self.elems
            self._elems = elems
            self._elem_view = None
            self.invalidate_cache()
            # self.on_update(self)
            self.on_update('elements', elems, old_elems)
        def activate(self):
            return self.activator(self)

        class StyleWrapper: # proxy for style
            def __init__(self, style_dict, obj):
                self.base_dict = style_dict
                self.base_obj = obj
            def __repr__(self):
                return "{}({})".format(type(self).__name__, self.base_dict)
            def __getitem__(self, item):
                return self.base_dict[item]
            def __setitem__(self, key, value):
                self.base_dict = dict(self.base_dict, **{key:value})
                self.base_obj.style = self.base_dict
            def __iter__(self):
                return iter(self.base_dict)
            def items(self):
                return self.base_dict.items()
            def keys(self):
                return self.base_dict.keys()
            def values(self):
                return self.base_dict.values()

        @property
        def style(self):
            if 'style' in self._attrs:
                return self.StyleWrapper(self._attrs['style'], self)
        @style.setter
        def style(self, styles):
            self['style'] = styles

        @property
        def class_list(self):
            if 'class' in self._attrs:
                return self.context.manage_class(self._attrs['class'])
            else:
                return []

        def invalidate_cache(self):
            if self._tree_cache is not None:
                self._tree_cache = None
                for p in tuple(self._parents):
                    p.invalidate_cache()
                    self._parents.remove(p)
        def __getitem__(self, item):
            if isinstance(item, str):
                item = item.replace("_", "-")
                return self._attrs[item]
            else:
                return self._elems[item]
        def __setitem__(self, item, value):
            if isinstance(item, str):
                item = item.replace("_", "-")
                old_value = self._attrs.get(item, None)
                self._attrs[item] = self.context.sanitize_value(value)
                self._attr_view = None
            else:
                old_value = self._elems[item]
                self._elems[item] = value
                self._elem_view = None
            self.invalidate_cache()
            self.on_update('attribute', value, old_value, subkey=item)
        def insert(self, where, child):
            if where is None:
                where = len(self._elems)
            self._elems.insert(where, child)
            self._elem_view = None
            self.invalidate_cache()
            self.on_update('element', child, None, subkey=where)
        def append(self, child):
            self.insert(None, child)
        def __delitem__(self, item):
            if isinstance(item, str):
                item = item.replace("_", "-")
                old_value = self._attrs.get(item, None)
                try:
                    del self._attrs[item]
                except KeyError:
                    pass
                else:
                    self._attr_view = None
            else:
                old_value = self._elems[item]
                del self._elems[item]
                self._elem_view = None
            self.invalidate_cache()
            self.on_update('attribute' if isinstance(item, str) else 'element', None, old_value, subkey=item)

        atomic_types = (int, bool, float)
        @classmethod
        def construct_etree_element(cls, elem, root, parent=None, attr_converter=None):
            if isinstance(elem, cls.atomic_types):
                elem = str(elem)
            if hasattr(elem, 'to_tree'):
                elem.to_tree(root=root, parent=parent, attr_converter=attr_converter)
            elif hasattr(elem, 'modify'):
                elem.modify().to_tree(root=root, parent=parent, attr_converter=attr_converter)
            elif isinstance(elem, ElementTree.Element):
                root.append(elem)
            elif isinstance(elem, (str, int, float, CSS)):
                elem = str(elem)
                kids = list(root)
                if len(kids) > 0:
                    if kids[-1].tail is None:
                        kids[-1].tail = elem
                    else:
                        kids[-1].tail += "\n" + elem
                else:
                    root.text = elem
            elif hasattr(elem, 'to_widget'):
                raise ValueError(
                    "can't convert {} to pure HTML. It looks like a Jupyter widget so look for the appropriate `JHTML` subclass.".format(
                        elem))
            else:
                raise ValueError("don't know what to do with {}".format(elem))

        attr_converter = None
        @classmethod
        def construct_etree_attrs(cls, attrs, attr_converter=None):
            _copied = False
            if 'style' in attrs:
                styles = attrs['style']
                if hasattr(styles, 'items'):
                    styles = CSS(**styles)
                if hasattr(styles, 'tostring'):
                    if not _copied:
                        attrs = attrs.copy()
                        _copied = True
                    attrs['style'] = styles.tostring()
            if 'class' in attrs:
                if not isinstance(attrs['class'], str):
                    if not _copied:
                        attrs = attrs.copy()
                        _copied = True
                    try:
                        iter(attrs['class'])
                    except TypeError:
                        attrs['class'] = str(attrs['class'])
                    else:
                        attrs['class'] = " ".join(str(c) for c in attrs['class'])
                    if len(attrs['class']) == 0:
                        del attrs['class']
            if attr_converter is None:
                attr_converter = cls.attr_converter
            if attr_converter is not None:
                attrs = attr_converter(attrs)
            return attrs
        @property
        def tree(self):
            return self.to_tree()
        def to_tree(self, root=None, parent=None, attr_converter=None):
            if parent is not None:
                self._parents.add(parent)
            if self._tree_cache is None:
                if root is None:
                    root = ElementTree.Element('root')
                attrs = self.construct_etree_attrs(self.attrs, attr_converter=attr_converter)
                my_el = ElementTree.SubElement(root, self.tag, attrs)
                if all(isinstance(e, str) for e in self.elems):
                    my_el.text = "\n".join(self.elems)
                else:
                    for elem in self.elems:
                        self.construct_etree_element(elem, my_el, parent=self, attr_converter=attr_converter)
                self._tree_cache = my_el
            elif root is not None:
                if self._tree_cache not in root:
                    root.append(self._tree_cache)
            return self._tree_cache
        def to_json(self, root=None, parent=None, attr_converter=None):
            tree = self.to_tree(root=root, parent=parent, attr_converter=attr_converter)
            return self.context.xml_to_json(tree)
        @classmethod
        def _prettyify(cls, current, *, indent, riffle, parent=None, index=-1, depth=0):
            # lightly adapted from https://stackoverflow.com/a/65808327/5720002
            for i, node in enumerate(current):
                cls._prettyify(node, indent=indent, riffle=riffle, parent=current, index=i, depth=depth + 1)
            if current.text is not None and len(current.text.strip()) > 0:
                current.text = (
                        riffle + textwrap.indent(current.text, prefix=indent * (depth+1))
                        + riffle + (indent * depth)
                )
            if parent is not None:
                if index == 0:
                    txt = parent.text
                    if txt is None:
                        txt = ""
                    parent.text = txt + riffle + (indent * depth)
                else:
                    txt = parent[index - 1].tail
                    if txt is not None:
                        txt = riffle + (indent * (depth)) + txt
                    else:
                        txt = ""
                    parent[index - 1].tail = txt + riffle + (indent * depth)
                if index == len(parent) - 1:
                    txt = current.tail
                    if txt is not None:
                        txt = riffle + (indent * (depth)) + txt
                    else:
                        txt = ""
                    current.tail = txt + riffle + (indent * (depth - 1))

        default_indent = "  "
        default_newline = "\n"
        def tostring(self, attr_converter=None, indent=None, method='html', riffle=True, prettify=False,
                     write_string=None,
                     **base_etree_opts):
            tree = self.to_tree(attr_converter=attr_converter)
            if prettify:
                if indent is not False:
                    if indent is None or indent is True:
                        indent = self.default_indent
                else:
                    indent = ""
                if riffle is not False:
                    if riffle is None or riffle is True:
                        riffle = self.default_newline
                else:
                    riffle = ""
                tree = copy.deepcopy(tree)
                self._prettyify(tree, indent=indent, riffle=riffle)
                if write_string is None:
                    write_string = ElementTree.tostring
                base_str = write_string(tree, **base_etree_opts)
            else:
                if indent is not None and indent is not False:
                    if indent is True:
                        indent = self.default_indent
                    tree = copy.deepcopy(tree)
                    ElementTree.indent(tree, space=indent)

                if riffle is not None and indent is not False:
                    if riffle is True:
                        riffle = self.default_newline
                    strs = [
                        s.decode() for s in ElementTree.tostringlist(
                            tree,
                            method=method,
                            **base_etree_opts
                        )
                    ]
                    base_str = riffle.join(strs)
                    if write_string is not None:
                        base_str = write_string(base_str)
                else:
                    if write_string is None:
                        write_string = ElementTree.tostring
                    base_str = write_string(tree)

            if hasattr(base_str, 'decode'):
                base_str = base_str.decode()
            return base_str

        def sanitize_key(self, key):
            key = key.replace("-", "_")
            for safe, danger in self.context.keyword_replacements.items():
                key = key.replace(danger, safe)
            return key
        def format(self, padding="", prefix="", linewidth=100):
            template_header = "{name}("
            template_footer = ")"
            template_pieces = []
            args_joiner = ", "
            full_joiner = ""
            elem_padding = ""
            def use_lines():
                nonlocal full_joiner, args_joiner, elem_padding, template_footer
                full_joiner = "\n"
                args_joiner = ",\n"
                elem_padding = padding + "  "
                template_footer = "{padding}  )"

            name = type(self).__name__
            tag = repr(self.tag) if len(template_pieces) > 1 else repr(self.tag)

            inner_comps = [
                (x.format(padding=padding+"  ", prefix=prefix, linewidth=linewidth) if isinstance(x, HTML.XMLElement) else repr(x))
                for x in self.elems if not (isinstance(x, str) and x.strip() == "")
            ]

            if not isinstance(self, HTML.TagElement):
                template_pieces.append("{tag}")
            if len(self.attrs) > 0:
                attr_pieces = ["{}={!r}".format(self.sanitize_key(k), v) for k,v in self.attrs.items()]
                if len(inner_comps) > 0:
                    template_pieces.append("{inner}")
                    template_pieces.append("{attrs}")
                else:
                    template_pieces.append("{attrs}")
            else:
                attr_pieces = []
                template_pieces.append("{inner}")
                # if "\n" not in inner:
                #     inner = inner.strip()
                #     full_joiner = ""
                #     template_footer = ")"

            template = full_joiner.join([
                template_header,
                args_joiner.join(template_pieces),
                template_footer
            ])
            out = template.format(
                padding=padding,
                tag=padding + "  " + tag,
                name=prefix+name,
                inner=args_joiner.join(elem_padding+x for x in inner_comps),
                attrs=args_joiner.join(elem_padding+x for x in attr_pieces),
            )
            if len(out) > linewidth + len(prefix):
                use_lines()
                template = full_joiner.join([
                    template_header,
                    args_joiner.join(template_pieces),
                    template_footer
                ])
                out = template.format(
                    padding=padding,
                    tag=padding + "  " + tag,
                    name=prefix + name,
                    inner=args_joiner.join(elem_padding + x for x in inner_comps),
                    attrs=args_joiner.join(elem_padding + x for x in attr_pieces),
                )
            return out
        def dump(self, prefix="", linewidth=80):
            print(self.format(prefix=prefix, linewidth=linewidth))
        def write(self, file, **opts):
            def write_str(tree, **base_opts):
                if isinstance(tree, str):
                    if hasattr(file, 'write'):
                        file.write(tree)
                    else:
                        with open(file, 'w+') as f:
                            f.write(tree)
                else:
                    tree.write(file, **base_opts)
            return self.tostring(write_string=write_str, **opts)
        MAX_REPR_LENGTH = 1000
        def __repr__(self):
            base_repr = "{}({}, {})".format(type(self).__name__, self.elems, self.attrs)
            if len(base_repr) > self.MAX_REPR_LENGTH + 3:
                split_len = self.MAX_REPR_LENGTH // 2
                base_repr = base_repr[:split_len] + "..." + base_repr[-split_len:]
            return base_repr
        def _repr_html_(self):
            return self.tostring()
        def _ipython_display_(self):
            self.display()
        def get_display_element(self):
            return HTML.Div(self, cls='jhtml')
        def get_mime_bundle(self):
            # from .WidgetTools import JupyterAPIs
            # display = JupyterAPIs.get_display_api()
            # from IPython.display import HTML as dispHTML
            wrapper = self.get_display_element()
            data = {
                'text/html': wrapper.tostring()
            }
            return data

        def _cross_plat_open(self, file, delay=5):
            import os, sys, subprocess, time
            if sys.platform.startswith('darwin'):  # macOS
                subprocess.run(['open', file])
            elif sys.platform.startswith('win'):  # Windows
                os.startfile(file, 'open')
            elif sys.platform.startswith('linux'):  # Linux
                subprocess.run(['xdg-open', file])
            else:
                raise NotImplementedError(f"unsure how to open file on {sys.platform}")
            time.sleep(delay)

        def display_in_browser(self):
            import tempfile as tf

            if self.tag.lower() != 'html':
                if self.tag.lower() != 'body':
                    wrapper = HTML.Body(self.get_display_element())
                else:
                    wrapper = self
                wrapper = HTML.Html(wrapper)
            else:
                wrapper = self

            with tf.NamedTemporaryFile(suffix='.html', prefix=type(self).__name__+"-", mode='w+',
                                       # delete=False
                                       ) as tmp_html:
                tmp_html.write(wrapper.tostring())
                tmp_html.seek(0)
                tmp_html.flush()
                self._cross_plat_open(tmp_html.name)

        def display_ipython(self):
            from .WidgetTools import JupyterAPIs

            display = JupyterAPIs.get_display_api()
            wrapper = self.get_display_element()
            return display.display(display.HTML(wrapper.tostring()))

        def display(self):
            from .WidgetTools import JupyterAPIs

            use_ipython = JupyterAPIs.in_jupyter_environment()
            if use_ipython:
                self.display_ipython()
            else:
                self.display_in_browser()
        @mixedmethod
        def _ipython_pinfo_(self):
            from ...Docs import jdoc
            return jdoc(self)

        def make_class_list(self):
            self._attrs['class'] = self._attrs['class'].split()
        def add_class(self, *cls, copy=True):
            return self.context.ClassAdder(self, cls, copy=copy).modify()
        def remove_class(self, *cls, copy=True):
            return self.context.ClassRemover(self, cls, copy=copy).modify()
        def add_styles(self, copy=True, **sty):
            return self.context.StyleAdder(self, copy=copy, **sty).modify()
        def remove_styles(self, copy=True, **sty):
            return self.context.StyleRemover(self, copy=copy, **sty).modify()
        # def remove_styles(self, copy=True, **sty):
        #     return HTML.StyleAdder(self, copy=copy, **sty).modify()

        def _find_child_node(self, etree):
            from collections import deque
            # BFS to try to find the element that matches
            remaining = deque()
            remaining.append(self)
            while remaining:
                elem = remaining.popleft()
                if isinstance(elem, self.context.ElementModifier):
                    elem = elem.modify()
                if isinstance(elem, HTML.XMLElement):
                    if etree == elem.tree:
                        return elem
                    else:
                        for e in elem.elems:
                            remaining.append(e)

        def find(self, path, find_element=True):
            base = self.tree.find(path)
            if find_element and base is not None:
                new = self._find_child_node(base)
                if new is not None:
                    base = new
            return base
        def findall(self, path, find_element=True):
            bases = self.tree.findall(path)
            if find_element:
                new = []
                for b in bases:
                    newb = self._find_child_node(b)
                    if newb is not None:
                        new.append(newb)
                    else:
                        new.append(b)
                bases = new
            return bases
        def iterfind(self, path, find_element=True):
            bases = self.tree.iterfind(path)
            for b in bases:
                if find_element:
                    newb = self._find_child_node(b)
                    if newb is not None:
                        yield newb
                    else:
                        yield b
                else:
                    yield b
        def _build_single_selector(self, root='.//', node_type='*', parents=None, **attrs):
            return "{root}{node_type}{atts}{parents}".format(
                root=root,
                node_type=node_type,
                atts='[' +
                     ' and '.join(
                         "@{k}='{v}'".format(k='class' if k == 'cls' else k, v=v)
                         for k,v in attrs.items()
                     ) + ']' if len(attrs) > 0 else '',
                parents="" if parents is None else ('.' + '.' * parents)
            )

        def _build_xpath_selector(self,  root='.//', node_type='*', parents=None, **attrs):
            attrs.update({
                'root':root,
                'node_type':node_type,
                'parents':parents
            })
            direct_prod_attrs = [
                [v]
                    if v is None else
                [str(vv) for vv in v]
                    if not isinstance(v, (str, int, float, bool)) else
                [str(v)]
                for v in attrs.values()
            ]
            selectors = [
                self._build_single_selector(
                    **dict(zip(attrs.keys(), p))
                )
                for p in itertools.product(*direct_prod_attrs)
            ]
            return " | ".join(selectors)

        def find_by_id(self, id, mode='first', parent=None, find_element=True):
            fn = {
                'first':self.find,
                'all':self.findall,
                'iter':self.iterfind
            }[mode]
            sel = ".//*[@id='{id}']{parents}".format(id=id, parents="" if parent is None else ('.'+'.'*parent))
            return fn(sel, find_element=find_element)
        def find_by_attributes(self,
                               *,
                               root='.//', node_type='*', parents=None,
                               mode='first',
                               find_element=True,
                               **attrs
        ):
            fn = {
                'first': self.find,
                'all': self.findall,
                'iter': self.iterfind
            }[mode]
            sel = self._build_xpath_selector(root=root, node_type=node_type, parents=parents, **attrs)
            return fn(sel, find_element=find_element)
        def build_selector(self, *dicts, **attrs):
            if len(dicts) == 0:
                return self._build_xpath_selector(**attrs)
            elif len(attrs) == 0:
                cur = self._build_xpath_selector(**dicts[0])
                for d in dicts[1:]:
                    if 'root' in d:
                        raise ValueError("root is inherited from previous selector")
                    d = dict(d, root=cur+'/')
                    cur = self._build_xpath_selector(**d)
                return cur
            else:
                raise ValueError("unsure what to do when given both dicts and kwargs?")


        def copy(self):
            import copy
            base = copy.copy(self)
            base.attrs = base.attrs.copy()
            base._tree_cache = None
            base._parents = weakref.WeakSet()
            return base

    base_element = XMLElement

    class Comment(XMLElement):
        def __init__(self, *elems, **attrs):
            super().__init__(ElementTree.Comment, *elems, **attrs)

    class TagElement(XMLElement):
        tag = None
        def __init__(self, *elems, **attrs):
            super().__init__(self.tag, *elems, **attrs)
        def __call__(self, *elems, **kwargs):
            return type(self)(
                self._elems + list(elems),
                activator=self.activator,
                on_update=self.on_update,
                **dict(self.attrs, **kwargs)
            )
    class Nav(TagElement): tag='nav'
    class Anchor(TagElement): tag='a'
    class Text(TagElement): tag='p'
    class Div(TagElement): tag='div'
    class Heading(TagElement): tag='h1'
    class SubHeading(TagElement): tag='h2'
    class SubsubHeading(TagElement): tag='h3'
    class SubsubsubHeading(TagElement): tag='h4'
    class SubHeading5(TagElement): tag='h5'
    class SubHeading6(TagElement): tag='h6'
    class Small(TagElement): tag='small'
    class Bold(TagElement): tag='b'
    class Italic(TagElement): tag='i'
    class Image(TagElement): tag='img'
    class ListItem(TagElement): tag='li'
    class BaseList(TagElement):
        def __init__(self, *elems, item_attributes=None, **attrs):
            if item_attributes is None:
                item_attributes = {}
            # elems = [HTML.ListItem(x, **item_attributes) if not isinstance(x, HTML.ListItem) else x for x in elems]
            super().__init__(*elems, **attrs)
    class List(BaseList): tag='ul'
    class NumberedList(BaseList): tag='ol'
    class Pre(TagElement): tag='pre'
    class Style(TagElement): tag='style'
    class Script(TagElement): tag='script'
    class Span(TagElement): tag='span'
    class Button(TagElement): tag='button'
    class TableRow(TagElement): tag='tr'
    class TableHeading(TagElement): tag='th'
    class TableHeader(TagElement): tag='thead'
    class TableFooter(TagElement): tag='tfoot'
    class TableBody(TagElement): tag='tbody'
    class TableItem(TagElement): tag='td'
    class Table(TagElement):
        tag = 'table'
        def __init__(self, *rows, headers=None, **attrs):
            if len(rows) == 1 and isinstance(rows[0], (list, tuple)):
                rows = rows[0]
            rows = [
                HTML.TableRow(
                    [HTML.TableItem(y) if not isinstance(y, HTML.TableItem) else y for y in x]
                ) if not isinstance(x, HTML.TableRow) else x for x in rows
            ]
            if headers is not None:
                rows = [
                    HTML.TableRow([HTML.TableHeading(x) if not isinstance(x, HTML.TableHeading) else x for x in headers])
                ] + rows
            super().__init__(rows, **attrs)

    class Canvas(TagElement): tag='canvas'

    A = Anchor
    class Abbr(TagElement): tag= "abbr"
    class Address(TagElement): tag= "address"
    class Area(TagElement): tag= "area"
    class Article(TagElement): tag= "article"
    class Aside(TagElement): tag= "aside"
    class Audio(TagElement): tag= "audio"
    class B(TagElement): tag= "b"
    class Base(TagElement): tag= "base"
    class Bdi(TagElement): tag= "bdi"
    class Bdo(TagElement): tag= "bdo"
    class Blockquote(TagElement): tag= "blockquote"
    class Body(TagElement): tag= "body"
    class Br(TagElement): tag= "br"
    class Caption(TagElement): tag= "caption"
    class Cite(TagElement): tag= "cite"
    class Code(TagElement): tag= "code"
    class Col(TagElement): tag= "col"
    class Colgroup(TagElement): tag= "colgroup"
    class Data(TagElement): tag= "data"
    class Datalist(TagElement): tag= "datalist"
    class Dd(TagElement): tag= "dd"
    class Del(TagElement): tag= "del"
    class Details(TagElement): tag= "details"
    class Dfn(TagElement): tag= "dfn"
    class Dialog(TagElement): tag= "dialog"
    class Dl(TagElement): tag= "dl"
    class Dt(TagElement): tag= "dt"
    class Em(TagElement): tag= "em"
    class Embed(TagElement): tag= "embed"
    class Fieldset(TagElement): tag= "fieldset"
    class Figcaption(TagElement): tag= "figcaption"
    class Figure(TagElement): tag= "figure"
    class Footer(TagElement): tag= "footer"
    class Form(TagElement): tag= "form"
    class Head(TagElement): tag= "head"
    class Header(TagElement): tag= "header"
    class Hr(TagElement): tag= "hr"
    class Html(TagElement): tag = "Html"
    i = Italic
    class Iframe(TagElement): tag= "iframe"
    Img = Image
    class Input(TagElement): tag= "input"
    class Ins(TagElement): tag= "ins"
    class Kbd(TagElement): tag= "kbd"
    class Label(TagElement): tag= "label"
    class Legend(TagElement): tag= "legend"
    Li = ListItem
    class Link(TagElement): tag= "link"
    class Main(TagElement): tag= "main"
    class Map(TagElement): tag= "map"
    class Mark(TagElement): tag= "mark"
    class Meta(TagElement): tag= "meta"
    class Meter(TagElement): tag= "meter"
    class Noscript(TagElement): tag= "noscript"
    class Object(TagElement): tag= "object"
    Ol = NumberedList
    P = Text
    class Optgroup(TagElement): tag= "optgroup"
    class Option(TagElement): tag= "option"
    class Output(TagElement): tag= "output"
    class Param(TagElement): tag= "param"
    class Picture(TagElement): tag= "picture"
    class Progress(TagElement): tag= "progress"
    class Q(TagElement): tag= "q"
    class Rp(TagElement): tag= "rp"
    class Rt(TagElement): tag= "rt"
    class Ruby(TagElement): tag= "ruby"
    class S(TagElement): tag= "s"
    class Samp(TagElement): tag= "samp"
    class Section(TagElement): tag= "section"
    class Select(TagElement): tag= "select"
    class Source(TagElement): tag= "source"
    class Strong(TagElement): tag= "strong"
    class Sub(TagElement): tag= "sub"
    class Summary(TagElement): tag= "summary"
    class Sup(TagElement): tag= "sup"
    class Svg(TagElement): tag= "svg"
    Tbody = TableBody
    Td = TableItem
    class Template(TagElement): tag= "template"
    class Textarea(TagElement): tag= "textarea"
    Tfoot = TableFooter
    Th = TableHeading
    Thead = TableHeader
    class Time(TagElement): tag= "time"
    class Title(TagElement): tag= "title"
    Tr = TableRow
    class Track(TagElement): tag= "track"
    class U(TagElement): tag= "u"
    Ul = List
    class Var(TagElement): tag= "var"
    class Video(TagElement): tag= "video"
    class Wbr(TagElement): tag= "wbr"

    # @classmethod
    # def extract_body(cls, etree, strip=True):
    #     text = etree.text
    #     if text is not None:
    #         if isinstance(text, str):
    #             text = [text]
    #     else:
    #         text = []
    #     tail = etree.tail
    #     if tail is not None:
    #         if isinstance(tail, str):
    #             tail = [tail]
    #     else:
    #         tail = []
    #     tag = etree.tag
    #
    #     elems = (
    #             [t.strip() if strip else t for t in text]
    #             + children
    #             + [t.strip() if strip else t for t in tail]
    #     )
    #     if strip:
    #         elems = [e for e in elems if not isinstance(e, str) or len(e) > 0]

class ContentXML(XMLBase):

    class Element(HTML.XMLElement):
        ignored_styles = CSS.known_properties
        def get_display_element(self):
            return HTML.Pre(self.tostring()).get_display_element()
        def tostring(self, method='xml', prettify=True, **opts):
            return super().tostring(method=method, prettify=prettify, **opts)

    base_element = Element
    class TagElement(Element):
        tag = None

        def __init__(self, *elems, **attrs):
            super().__init__(self.tag, *elems, **attrs)

        def __call__(self, *elems, **kwargs):
            return type(self)(
                self._elems + list(elems),
                activator=self.activator,
                on_update=self.on_update,
                **dict(self.attrs, **kwargs)
            )

    class DeclarativeElement(TagElement):
        def __init__(self, *elems, **attrs):
            self.tag = type(self).__name__
            super().__init__(*elems, **attrs)

    class Comment(HTML.Comment): ...

    class PrefixedElement(Element):
        prefix = None
        def __init__(self, base_tag, *elems, **attrs):
            self.base_tag = base_tag
            super().__init__(self.prefix + base_tag, *elems, **attrs)
        def __call__(self, *elems, **kwargs):
            return type(self)(
                self.base_tag,
                self._elems + list(elems),
                activator=self.activator,
                on_update=self.on_update,
                **dict(self.attrs, **kwargs)
            )





from .CoordinateSystemConverter import CoordinateSystemConverter
from .CommonCoordinateSystems import CartesianCoordinates3D, ZMatrixCoordinates
from ...Numputils import vec_norms, vec_angles, pts_dihedrals, dist_deriv, angle_deriv, dihed_deriv
from ... import Numputils as nput
import numpy as np
# this import gets bound at load time, so unfortunately PyCharm can't know just yet
# what properties its class will have and will try to claim that the files don't exist

class CartesianToZMatrixConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to ZMatrix coordinates
    """

    @property
    def types(self):
        return (CartesianCoordinates3D, ZMatrixCoordinates)

    @staticmethod
    def get_dists(points, centers):
        return vec_norms(centers-points)
    @staticmethod
    def get_angles(lefts, centers, rights):
        # need to look up again what the convention is for which atom is the central one...
        v1s = lefts-centers
        v2s = rights-centers
        return vec_angles(v1s, v2s)[0]
    @staticmethod
    def get_diheds(points, centers, seconds, thirds):
        return pts_dihedrals(points, centers, seconds, thirds)

    def convert_many(self, coords, *, ordering, use_rad=True, return_derivs=False, **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """
        #TODO: introduce fast derivs back into this coordinate system by extracting "specs" from the ordering
        if ordering is None:
            raise NotImplementedError("an ordering must be supplied")
            ordering = range(len(coords[0]))
        base_shape = coords.shape
        new_coords = np.reshape(coords, (np.prod(base_shape[:-1]),) + base_shape[-1:])
        new_coords, ops = self.convert(new_coords,
                                       ordering=ordering,
                                       use_rad=use_rad,
                                       return_derivs=return_derivs,
                                       **kw)
        single_coord_shape = (base_shape[-2]-1, new_coords.shape[-1])
        new_shape = base_shape[:-2] + single_coord_shape
        new_coords = np.reshape(new_coords, new_shape)
        if return_derivs:
            ders = ops['derivs']
            # we assume we get a list of derivatives?
            reshaped_ders = [None]*len(ders)
            for i, v in enumerate(ders):
                single_base_shape = (base_shape[-2], new_coords.shape[-1])
                ders_shape = coords.shape + single_base_shape*i + single_coord_shape
                v = v.reshape(ders_shape)
                reshaped_ders[i] = v
                ops['derivs'] = reshaped_ders
        return new_coords, ops

    def convert(self, coords, *, ordering, use_rad=True, return_derivs=None, order=None,
                strip_embedding=False,
                **kw):
        """The ordering should be specified like:

        [
            [n1],
            [n2, n1]
            [n3, n1/n2, n1/n2]
            [n4, n1/n2/n3, n1/n2/n3, n1/n2/n3]
            [n5, ...]
            ...
        ]

        :param coords:    array of cartesian coordinates
        :type coords:     np.ndarray
        :param use_rad:   whether to user radians or not
        :type use_rad:    bool
        :param ordering:  optional ordering parameter for the z-matrix
        :type ordering:   None or tuple of ints or tuple of tuple of ints
        :param kw:        ignored key-word arguments
        :type kw:
        :return: z-matrix coords
        :rtype: np.ndarray
        """


        if return_derivs is None and order is not None and order > 0:
            return_derivs = True

        ncoords = len(coords)
        orig_ol = ZMatrixCoordinates.canonicalize_order_list(ncoords, ordering)
        ol = orig_ol
        nol = len(ol)
        ncol = len(ol[0])
        fsteps = ncoords / nol
        steps = int(round(fsteps))

        # print(">> c2z >> ordering:", ol)

        multiconfig = nol < ncoords
        if multiconfig:
            ol = ZMatrixCoordinates.tile_order_list(ol, ncoords)
            mc_ol = ol.copy()

        # we define an order map that we'll index into to get the new indices for a
        # given coordinate
        om = 1 + np.argsort(ol[:, 0])
        if return_derivs:
            og_ord = None
            if order is None:
                order = 1
            elif isinstance(order, list):
                og_ord = order
                order = max(order)

            if strip_embedding:
                specs = [
                    p
                    for n,(i, j, k, l) in enumerate(ordering[1:])
                    for p in (
                            [(i, j), (i, j, k), (i, j, k, l)]
                                if n > 1 else
                            [(i, j), (i, j, k)]
                                if n > 0 else
                            [(i,j)]
                    )
                ]
                if len(specs) > 3:
                    ix = np.concatenate([[0, 1], np.arange(3, len(specs), 3)])
                    jx = np.concatenate([[2], np.arange(4, len(specs), 3)])
                    kx = np.arange(5, len(specs), 3)
                elif len(specs) > 1:
                    ix = np.array([0, 1])
                    jx = np.array([2])
                    kx = None
                else:
                    ix = np.array([0])
                    jx = None
                    kx = None
            else:
                specs = [
                    p
                    for n, (i, j, k, l) in enumerate(ordering[1:])
                    for p in (
                        [(i, j), (i, j, k), (i, j, k, l)]
                    )
                ]
                ix = np.arange(0, len(specs), 3)
                jx = np.arange(1, len(specs), 3)
                kx = np.arange(2, len(specs), 3)

            crd = coords.reshape((steps, nol, 3))
            base_conv = nput.internal_coordinate_tensors(crd, specs, order=order, angle_ordering='ijk')
            if strip_embedding:
                dists = base_conv[0][..., ix]
                if jx is None:
                    angles = np.zeros((len(base_conv), 1))
                else:
                    angles = np.array([
                        np.concatenate([[0], a])
                        for a in base_conv[0][..., jx]
                    ])
                if kx is None:
                    diheds = np.zeros((len(base_conv), 2))
                else:
                    diheds = np.array([
                        np.concatenate([[0], d])
                        for d in base_conv[0][..., kx]
                    ])
            else:
                dists = base_conv[0][..., ix]
                angles = base_conv[0][..., jx]
                diheds = base_conv[0][..., kx]


            base_derivs = base_conv[1:]
            derivs = []
            nats = nol
            for n,d in enumerate(base_derivs):
                deriv = np.zeros(
                    (steps,)
                    + ((nats)*3,)*(n+1)
                    + ((nats-1) * 3,)
                )
                if strip_embedding:
                    deriv[..., np.arange(0, deriv.shape[-1], 3)] = d[..., ix]
                    if jx is not None:
                        deriv[..., np.arange(1, deriv.shape[-1], 3)] = np.pad(
                            d[..., jx],
                            [[0, 0]] * (d.ndim - 1) + [[1, 0]]
                        )
                    if kx is not None:
                        deriv[..., np.arange(2, deriv.shape[-1], 3)] = np.pad(
                            d[..., kx],
                            [[0, 0]] * (d.ndim - 1) + [[2, 0]]
                        )
                else:
                    deriv[..., np.arange(0, deriv.shape[-1], 3)] = d[..., ix]
                    if jx is not None:
                        deriv[..., np.arange(1, deriv.shape[-1], 3)] = d[..., jx]
                    if kx is not None:
                        deriv[..., np.arange(2, deriv.shape[-1], 3)] = d[..., kx]
                derivs.append(deriv)

            if og_ord is not None:
                derivs = [derivs[o-1] for o in og_ord]
        else:
            # need to check against the cases of like 1, 2, 3 atom molecules
            # annoying but not hard
            if return_derivs:
                derivs = [
                    np.zeros(coords.shape + (nol-1, 3)),
                    np.zeros(coords.shape + (nol, 3) + (nol - 1, 3))
                ]
            if not multiconfig:
                ix = ol[1:, 0]
                jx = ol[1:, 1]
                dists = self.get_dists(coords[ix], coords[jx])
                if return_derivs:
                    _dists, dist_derivs, dist_derivs_2 = dist_deriv(coords, ix, jx, order=2)
                    drang = np.arange(len(ix))
                    derivs[0][ix, :, drang, 0] = dist_derivs[0]
                    derivs[0][jx, :, drang, 0] = dist_derivs[1]

                    for i, x1 in enumerate([ix, jx]):
                        for j, x2 in enumerate([ix, jx]):
                            # print(i, j, x1, x2,
                            #       # dist_derivs_2[i, j][0, 0],
                            #       drang
                            #       )
                            derivs[1][x1, :, x2, :, drang, 0] = dist_derivs_2[i, j]

                if len(ol) > 2:
                    ix = ol[2:, 0]
                    jx = ol[2:, 1]
                    kx = ol[2:, 2]
                    angles = np.concatenate( (
                        [0], self.get_angles(coords[ix], coords[jx], coords[kx])
                    ) )
                    if not use_rad:
                        angles = np.rad2deg(angles)
                    if return_derivs:
                        _angles, angle_derivs, angle_derivs_2 = angle_deriv(coords, jx, ix, kx, order=2)
                        drang = 1+np.arange(len(ix))
                        # print(">>>>", np.max(np.abs(angle_derivs)))
                        derivs[0][jx, :, drang, 1] = angle_derivs[0]
                        derivs[0][ix, :, drang, 1] = angle_derivs[1]
                        derivs[0][kx, :, drang, 1] = angle_derivs[2]

                        for i, x1 in enumerate([jx, ix, kx]):
                            for j, x2 in enumerate([jx, ix, kx]):
                                derivs[1][x1, :, x2, :, drang, 1] = angle_derivs_2[i, j]
                else:
                    angles = np.array([0.])
                if len(ol) > 3:
                    ix = ol[3:, 0]
                    jx = ol[3:, 1]
                    kx = ol[3:, 2]
                    lx = ol[3:, 3]
                    if ol.shape[1] == 5:
                        raise NotImplementedError("psi angles might be unnecessary")
                        ix = ix.copy()
                        jx = jx.copy()
                        kx = kx.copy()
                        lx = lx.copy()
                        fx = ol[3:, 4]
                        swap_pos = np.where(fx == 1)
                        swap_i = ix[swap_pos]
                        swap_j = jx[swap_pos]
                        swap_k = kx[swap_pos]
                        swap_l = lx[swap_pos]
                        ix[swap_pos] = swap_l
                        jx[swap_pos] = swap_i
                        kx[swap_pos] = swap_j
                        lx[swap_pos] = swap_k

                    diheds = np.concatenate( (
                        [0, 0],
                        self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                    ) )
                    if not use_rad:
                        diheds = np.rad2deg(diheds)
                    if return_derivs:
                        _diheds, dihed_derivs, dihed_derivs_2 = dihed_deriv(coords, ix, jx, kx, lx, order=2)
                        drang = 2+np.arange(len(ix))
                        derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                        derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                        derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                        derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                        for i, x1 in enumerate([ix, jx, kx, lx]):
                            for j, x2 in enumerate([ix, jx, kx, lx]):
                                derivs[1][x1, :, x2, :, drang, 2] = dihed_derivs_2[i, j]
                else:
                    diheds = np.array([0, 0])
                ol = ol[1:]

            else: # multiconfig

                # we do all of this stuff with masking operations in the multiconfiguration cases
                mask = np.repeat(True, ncoords)
                mask[np.arange(0, ncoords, nol)] = False
                ix = ol[mask, 0]
                jx = ol[mask, 1]
                dists = self.get_dists(coords[ix], coords[jx])
                if return_derivs:
                    _, dist_derivs, dist_derivs_2 = dist_deriv(coords, ix, jx, order=2)
                    drang = np.arange(nol-1)
                    nreps = int(len(ix)/(nol-1))
                    drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                    derivs[0][ix, :, drang, 0] = dist_derivs[0]
                    derivs[0][jx, :, drang, 0] = dist_derivs[1]

                    for i, x1 in enumerate([ix, jx]):
                        for j, x2 in enumerate([ix, jx]):
                            derivs[1][x1, :, x2 % nol, :, drang, 0] = dist_derivs_2[i, j]

                if nol>2:
                    # set up the mask to drop all of the first bits
                    mask[np.arange(1, ncoords, nol)] = False
                    ix = ol[mask, 0]
                    jx = ol[mask, 1]
                    kx = ol[mask, 2]
                    angles = self.get_angles(coords[ix], coords[jx], coords[kx])
                    angles = np.append(angles, np.zeros(steps))
                    insert_pos = np.arange(0, ncoords-1*steps-1, nol-2)
                    angles = np.insert(angles, insert_pos, 0)
                    angles = angles[:ncoords-steps]
                    if not use_rad:
                        angles = np.rad2deg(angles)
                    if return_derivs:
                        # we might need to mess with the masks akin to the insert call...
                        _, angle_derivs, angle_derivs_2 = angle_deriv(coords, jx, ix, kx, order=2)
                        drang = 1+np.arange(nol-2)
                        nreps = int(len(ix)/(nol-2))
                        drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                        derivs[0][jx, :, drang, 1] = angle_derivs[0]
                        derivs[0][ix, :, drang, 1] = angle_derivs[1]
                        derivs[0][kx, :, drang, 1] = angle_derivs[2]

                        for i, x1 in enumerate([ix, jx, kx]):
                            for j, x2 in enumerate([ix, jx, kx]):
                                derivs[1][x1, :, x2 % nol, :, drang, 0] = angle_derivs_2[i, j]
                else:
                    angles = np.zeros(ncoords-steps)

                if nol > 3:
                    # set up mask to drop all of the second atom bits (wtf it means 'second')
                    mask[np.arange(2, ncoords, nol)] = False
                    ix = ol[mask, 0]
                    jx = ol[mask, 1]
                    kx = ol[mask, 2]
                    lx = ol[mask, 3]
                    if ol.shape[1] == 5:
                        raise ValueError("Unclear if there is a difference between tau and psi")
                        ix = ix.copy()
                        jx = jx.copy()
                        kx = kx.copy()
                        lx = lx.copy()
                        fx = ol[mask, 4]
                        swap_pos = np.where(fx == 1)
                        swap_i = ix[swap_pos]
                        swap_j = jx[swap_pos]
                        swap_k = kx[swap_pos]
                        swap_l = lx[swap_pos]
                        ix[swap_pos] = swap_l
                        jx[swap_pos] = swap_i
                        kx[swap_pos] = swap_j
                        lx[swap_pos] = swap_k
                    # print(ol)

                    diheds = self.get_diheds(coords[ix], coords[jx], coords[kx], coords[lx])
                    # pad diheds to be the size of ncoords
                    diheds = np.append(diheds, np.zeros(2*steps))

                    # insert zeros where undefined
                    diheds = np.insert(diheds, np.repeat(np.arange(0, ncoords-2*steps-1, nol-3), 2), 0)
                    # take only as many as actually used
                    diheds = diheds[:ncoords-steps]
                    if not use_rad:
                        diheds = np.rad2deg(diheds)
                    if return_derivs:
                        # Negative sign because my dihed_deriv code is for slightly different
                        # ordering than expected
                        _, dihed_derivs, dihed_derivs_2 = dihed_deriv(coords, ix, jx, kx, lx, order=2)
                        drang = 2+np.arange(nol-3)
                        nreps = int(len(ix)/(nol-3))
                        drang = np.broadcast_to(drang[np.newaxis], (nreps,) + drang.shape).flatten()
                        derivs[0][ix, :, drang, 2] = dihed_derivs[0]
                        derivs[0][jx, :, drang, 2] = dihed_derivs[1]
                        derivs[0][kx, :, drang, 2] = dihed_derivs[2]
                        derivs[0][lx, :, drang, 2] = dihed_derivs[3]

                        for i, x1 in enumerate([ix, jx, kx, lx]):
                            for j, x2 in enumerate([ix, jx, kx, lx]):
                                derivs[1][x1, :, x2 % nol, :, drang, 0] = dihed_derivs_2[i, j]

                else:
                    diheds = np.zeros(ncoords-steps)

                # after the np.insert calls we have the right number of final elements, but too many
                # ol and om elements and they're generally too large
                # so we need to shift them down and mask out the elements we don't want
                mask = np.repeat(True, ncoords)
                mask[np.arange(0, ncoords, nol)] = False
                ol = np.reshape(ol[mask], (steps, nol-1, ncol))-np.reshape(np.arange(steps), (steps, 1, 1))
                ol = np.reshape(ol, (ncoords-steps, ncol))
                om = np.reshape(om[mask], (steps, nol-1))-nol*np.reshape(np.arange(steps), (steps, 1))-1
                om = np.reshape(om, (ncoords-steps,))

        final_coords = np.array(
            [
                dists, angles, diheds
            ]
        ).T

        if multiconfig:
            # figure out what to use for the axes
            origins = coords[mc_ol[1::nol,  1]]
            x_axes  = coords[mc_ol[1::nol,  0]] - origins # the first displacement vector
            y_axes  = coords[mc_ol[2::nol,  0]] - origins # the second displacement vector (just defines the x-y plane, not the real y-axis)
            axes = np.array([x_axes, y_axes]).transpose((1, 0, 2))
        else:
            origins = coords[ol[0, 1]]
            axes = np.array([coords[ol[0, 0]] - origins, coords[ol[1, 0]] - origins])

        ol = orig_ol
        om = om - 1
        if ncol == 5:
            ordering = np.array([
                        np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]], ol[:, 4]
                    ]).T
        else:
            ordering = np.array([
                    np.argsort(ol[:, 0]), om[ol[:, 1]], om[ol[:, 2]], om[ol[:, 3]]
                ]).T
        opts = dict(use_rad=use_rad, ordering=ordering, origins=origins, axes=axes)

        # if we're returning derivs, we also need to make sure that they're ordered the same way the other data is...
        if return_derivs:
            opts['derivs'] = derivs#[:1]

        return final_coords, opts

__converters__ = [ CartesianToZMatrixConverter() ]


import sys, os
import tempfile as tf, io

__all__ = [
    "OutputRedirect",
    "DefaultDirectory"
]

class OutputRedirect:
    def __init__(self,
                 redirect=True,
                 stdout=None, stderr=None,
                 capture_output=False,
                 capture_errors=None,
                 file_handles=False
                 ):
        self.redirect = redirect
        if capture_errors is None:
            capture_errors = capture_output
        self.capture_output = capture_output
        self.capture_errors = capture_errors
        if capture_output:
            stdout = self.get_handle(stdout, file_handles)
        if capture_errors:
            stderr = self.get_handle(stderr, file_handles)
        self._tmp = [None, None]
        self._stdout = None
        self._stderr = None
        self._devnull = None
        self._out_stream = stdout
        self._err_stream = stderr

    @classmethod
    def get_handle(cls, handles=None, file_handles=False):
        if handles is not None:
            return handles
        if not file_handles:
            return io.StringIO()
        else:
            return None

    @classmethod
    def get_temp_stream(cls):
        return tf.NamedTemporaryFile(mode='w+').__enter__()

    def __enter__(self):
        if self.redirect:
            self._stdout = sys.stdout
            self._stderr = sys.stderr
            if (
                    not self.capture_output
                    and self._out_stream is None
            ) or (
                    not self.capture_errors
                    and self._err_stream is None
            ):
                self._devnull = open(os.devnull, 'w+').__enter__()

            if self._out_stream is None:
                if self.capture_output:
                    self._tmp[0] = self.get_temp_stream()
                    sys.stdout = self._tmp[0]
                else:
                    sys.stdout = self._devnull
            else:
                if isinstance(self._out_stream, str):
                    self._tmp[0] = open(self._out_stream, 'a').__enter__()
                    sys.stdout = self._tmp[0]
                else:
                    sys.stdout = self._out_stream

            if self._err_stream is None:
                if self.capture_errors:
                    self._tmp[1] = self.get_temp_stream()
                    sys.stderr = self._tmp[1]
                else:
                    sys.stderr = self._devnull
            else:
                if isinstance(self._err_stream, str):
                    self._tmp[1] = open(self._err_stream, 'a').__enter__()
                    sys.stderr = self._tmp[1]
                else:
                    sys.stderr = self._err_stream
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.redirect:
            sys.stdout = self._stdout
            sys.stderr = self._stderr
            if self._devnull is not None:
                self._devnull.__exit__(exc_type, exc_val, exc_tb)
                self._devnull = None
            self._stdout = None
            self._stderr = None

            if self._tmp[0] is not None:
                self._tmp[0].__exit__(exc_type, exc_val, exc_tb)
                self._tmp[0] = None
            if self._tmp[1] is not None:
                self._tmp[1].__exit__(exc_type, exc_val, exc_tb)
                self._tmp[1] = None

class DefaultDirectory:
    def __init__(self, output_dir=None, chdir=True, **tempdir_opts):
        self.chdir = chdir
        self._outdir = output_dir
        self._curdir = None
        self._tmp = None
        self._opts = tempdir_opts

    def get_temp_dir(self):
        return tf.TemporaryDirectory(**self._opts).__enter__()

    @property
    def dirname(self):
        if self._tmp is None:
            return None
        if isinstance(self._tmp, str):
            return self._tmp
        else:
            return self._tmp.name

    def __enter__(self):
        if self._outdir is None:
            self._tmp = self.get_temp_dir()
        else:
            self._tmp = self._outdir
        if self.chdir:
            self._curdir = os.getcwd()
            os.chdir(self.dirname)
        return self._tmp

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.chdir:
            os.chdir(self._curdir)
        if self._outdir is None:
            self._tmp.__exit__()
        self._tmp = None


"""
A growing package of assorted functionality that finds use across many different packages, but doesn't attempt to
provide a single unified interface for doing certain types of projects.

All of the McUtils packages stand mostly on their own, but there will be little calls into one another here and there,
especially pieces using `Numputils`

The more scientifically-focused `Psience` package makes significant use of `McUtils` as do various packages that have
been written over the years.
"""

__all__ = []

from . import Data
__all__ += ["Data"]
from . import Numputils
__all__ += ["Numputils"]
from . import ExternalPrograms
__all__ += ["ExternalPrograms"]
from . import Plots
__all__ += ["Plots"]
from . import Jupyter
__all__ += ["Jupyter"]
from . import Parsers
__all__ += ["Parsers"]
from . import Formatters
__all__ += ["Formatters"]
from . import Coordinerds
__all__ += ["Coordinerds"]
from . import Zachary
__all__ += ["Zachary"]
from . import GaussianInterface
__all__ += ["GaussianInterface"]
from . import Extensions
__all__ += ["Extensions"]
from . import Scaffolding
__all__ += ["Scaffolding"]
from . import Parallelizers
__all__ += ["Parallelizers"]
from . import Devutils
__all__ += ["Devutils"]
from . import Docs
__all__ += ["Docs"]
from . import Misc
__all__ += ["Misc"]